package edu.wpi.cs.dsrg.utils;

/////////////////////////// COPYRIGHT /////////////////////////////////////
//
//  Rainbow 1.0 - XQuery Processing.
//
//  Copyright 2003, All Rights Reserved.
//
//  Permission to use, copy, modify and distribute this software and
//  its documentation for educational, research and non-profit
//  purposes, without fee, and without a written agreement is hereby
//  granted, provided that this copyright notice appear in all copies.
//
//  Permission to incorporate this software into commercial products
//  may be obtained directly from the authors.
//
//  This software is provided "as is" without express or implied warranty.
//  In no event will the authors be held liable for any damages arising
//  from the use of this software.
//
//  Authors:
//  Steffen Christ, Katica Dimitrova, Luping Ding, Michal A. Dobos,
//  Maged El-Sayed, Geoff Greene, Huzefa Hakimi, Justin Lutz, Guy D. Mann,
//  Mukesh Mulchandani, Brian Murphy, Bradford Pielech, Alex Proschitsky,
//  Elke A. Rundensteiner, Ling Wang, Song Wang, and Xin Zhang
//
//      Computer Science Department
//      Worcester Polytechnic Institute
//      Worcester, MA 01609 USA
//
//      http://davis.wpi.edu/~dsrg/rainbow/index.htm
//
//  Send bug reports, bug fixes, enhancements, requests, or comments to
//  the Rainbow Development Team: <rainbow@WPI.EDU>.
//
/////////////////////////// COPYRIGHT /////////////////////////////////////

import java.util.*;
import java.lang.*;
import java.io.PrintWriter;
import java.io.StringWriter;

import edu.wpi.cs.dsrg.xmldb.xat.common.dompattern.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.bindingtree.BindingTree;
import edu.wpi.cs.dsrg.xmldb.xat.common.treestructure.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.*;
import edu.wpi.cs.dsrg.xmldb.xat.common.operator.xmloperator.navigation.*;

/**
 * This class defines the BindingTable, which is used to represent variable binding within the tree.
 * @author  Rainbow Group, DSRG, WPI
 * @version RainbowCore 1.0
 * @since   1.0
 */
 
public class BindingTable extends java.util.Hashtable {

    /**
     * Used for prevent infinite recursive call of variable bindings.
     * Only used in getVariableValue(...).
     * @since 1.0
     */
    Hashtable VarHash = new Hashtable();
    /**
     * Used for prevent infinite recursive call.
     * Only used in getChain(...).
     * @since 1.0
     */
    Hashtable ChainHash = new Hashtable();
/**
 * The default BindingTable constructor.
 * @since 1.0
 */
public BindingTable() {
	super();
}
/**
 * The BindingTable constructor using given capacity.
 * @param initialCapacity The capacity of the BindingTable.
 * @since 1.0
 */
public BindingTable(int initialCapacity) {
	super(initialCapacity);
}
/**
 * The BindingTable constructor.
 * @param initialCapacity The capacity of the BindingTable.
 * @param loadFactor float
 * @since 1.0
 */
public BindingTable(int initialCapacity, float loadFactor) {
	super(initialCapacity, loadFactor);
}
/**
 * The BindingTable constructor.
 * @param t The Map for BindingTable.
 * @since 1.0
 */
public BindingTable(java.util.Map t) {
	super(t);
}
/**
 * Compute the full path of a given variable.
 * The full path will be simplified by the following logics:
 * 1) (_x_(E.p.)).p. = E.p.
 * 2) (_x_(_p_E)).p. = _p_E
 * 3) (_x_(E.q.)).p. = E.r., where .p., .q., and .r. follow following table:
 *    .p.		.q.		.r.
 *    'path.	.//.	.path.
 *    .path.	.*.		.path.
 *    .//.		.any.	.path..//.
 *    .*.		.any.	.path.
 * 4) if there are more than one items in the tagger, then based on the distributility:
 *	  _x_(E + F) = _x_E + _x_F
 *
 * @return NavigationStep[] The computed full path for a given virable.
 * @param Key The given variable computing path for.
 * @since 1.0
 */
public NavigationStep[] computeSimplifiedFullPath(String key) {
    /* first use the getVariableValue(String, true) to recursively get the full path.*/
    if (!this.containsKey(key))
        return null;
    // make a ArrayList object for this array for ease of manipulation.
    NavigationStep[] result = this.getVariableValue(key, true);

    // Make The BindingTree out of the value we have.
    BindingTree bindtree = new BindingTree(result);

    /* second, use the rules to reduce the binding tree. */
    bindtree.reduce();

    // convert it back to NavigationStep array.
    return bindtree.toNavigationStep();

}
/**
 * Check if a value exists in the hashtable
 * @return boolean Trur: The given key exists in hashtable. False: otherwise.
 * @param key The variable looking for.
 * @since 1.0
 */
public boolean contains(String key) {
    return super.contains(key);
}
/**
 * Get the variable binding of a given key.
 * @return Object The Binding value for a given key.
 * @param key The key of the variable binding looking for.
 * @deprecated You should use getVariableValue(...) instead.
 * Note: Disable the usage of the original get() outside of this class.
 */
public Object get(Object key) {
    return super.get(key);
}
/**
 * It will get all the VarStep alone this Key in the
 * binding table.
 * @return Vector A Vector of VarStep value, which indicate where the key coming from.
 * @param key The Key searching for.
 * @since 1.0
 */
public Vector getChain(String key) {
    this.ChainHash = new Hashtable();
    return getChainRecursive(key);
}
/**
 * Get all the VarStep alone this Key in the binding table.
 * @return Vector A Vector of VarStep value, which indicate where the key coming from.
 * @param key The Key searching for.
 * @since 1.0
 */
public Vector getChainRecursive(String key) {

    Vector resultVector = new Vector();

    if(key == null){
	    return resultVector;
    }

    // check for infinite loop.
    if(this.ChainHash.get(key) != null){
	    // found a circle call.
	    System.err.println("ERROR: BindingTable.getChainRecursive(String) infinite loop for key:" + key + "! Abort!");
	    // reset the VarHash.
	    this.ChainHash = new Hashtable();
	    // give back the unrecursive answer.
	    return resultVector;
    }else{
	    // put to the varhash.
	    this.ChainHash.put(key,key);
    }

    // first, get all of its steps.
    NavigationStep[] result = (NavigationStep[]) this.get(key);
    if (result == null){
    	// remove it from the chain hash.
    	this.ChainHash.remove(key);
        return resultVector;
    }

    // then, recursively add all the VarSteps's step values.
    for (int j = 0; j < result.length; j++) {
	    if(result[j] instanceof VarStep){
		    resultVector.add(result[j].getStepValue());
	        resultVector.addAll(getChainRecursive(result[j].getStepValue()));
	    }
    }

    // remove it from the chain hash.
	this.ChainHash.remove(key);
    return resultVector;
}
/**
 * Get the key by the variable binding.
 * @return String The key of given binding.
 * @param value The value searching for.
 * @since 1.0
 */
public String getKeyByValue(String value) {
	String keyValue = new String("");
	for(Enumeration keyEnum = this.keys(); keyEnum.hasMoreElements() ;){
		String key = (String)keyEnum.nextElement();
		NavigationStep[] valueStep = getVariableValue(key);
		String valueString = navStepsToString(valueStep);
		if(valueString.equals(value)){
			return key;
		}
	} 
	return null;
}
/**
 * Given an XML path, plus "/SOURCE", return the Key as a string
 * @return String The key of the given XML path.
 * @param value An XML path.
 * @since 1.0
 */
public String getKeyByXMLValue(String value) {

	for(Enumeration keyEnum = this.keys(); keyEnum.hasMoreElements() ;){
		String key = (String)keyEnum.nextElement();
		NavigationStep[] valueStep = getVariableValue(key);
		String valueString = simpleToXMLString(valueStep);
		if(valueString.equals(value)){
			return key;
		}
	}
	//if none found, return null
	return null;
}
/**
 * Given an XML path, plus "/SOURCE", return the Key as a string
 * If more than one mapping can exist, use the method that returns a vector 
 * @return String The vector containing all the keys of given XML value.
 * @param value The XML value searching for.
 * @since 1.0
 */
public Vector getKeysByXMLValue(String value) {
	Vector retValues = new Vector();
	for(Enumeration keyEnum = this.keys(); keyEnum.hasMoreElements() ;){
		String key = (String)keyEnum.nextElement();
		NavigationStep[] valueStep = getVariableValue(key);
		String valueString = simpleToXMLString(valueStep);
		if(valueString.equals(value)){
			retValues.add(key);
		}
	} 
	return retValues;
}
/**
 * Get the tag value from an open/close tag of an XML element.
 * The following cases will be handled:
 * 1. Open tag: <aaa>
 * 2. Open tag with attributes: <aaa b="cc">
 * 3. Close tag: </aaa>
 * 4. brief tag: <aaa/>
 * @return String The Tag value of a given XML element.
 * @param element The XML element searching for.
 * @since 1.0
 */
public static String getTagValue(String element) {
    if (element == null)
        return "";

    // locate the first position of /.
    int posSlash = element.indexOf("/");
    // locate the first position of <.
    int posLeftBracket = element.indexOf("<");
    // locate the first position of ' '.
    int posSpace = element.indexOf(" ");
    // locate the first position of '>'.
    int posRightBracket = element.indexOf(">");

    String tagValue = "";

    if (posLeftBracket == 0 && posRightBracket == element.length() - 1) {
        if (posSlash == -1) {
            // that's an open tag.
            if (posSpace == -1) {
                // with no attributes.
                tagValue = element.substring(1, element.length() - 1);
            } else {
                // with attributes.
                tagValue = element.substring(1, posSpace);
            }
        } else {
            // that's might be a close tag or a brief tag.
            if (posSlash == 1) {
                // that's a close tag.
                tagValue = element.substring(2, element.length() - 1);
            } else
                if (posSlash == posRightBracket - 1) {
                    // that's a abbrev tag.
                    tagValue = element.substring(1, element.length() - 2);
                } else {
                        System.err.println("Unknow type of tags!");
                }
        }
    } else
        if (posLeftBracket == 0 && posSpace != -1) {
            // this is a imcomplete open tag, with an attribute in it.
            tagValue = element.substring(1, posSpace);
        } else
            if (posLeftBracket == 0 && posSpace == -1 && element.length() > 1) {
                // this is a imcomplete open tag, with an attribute in it.
                tagValue = element.substring(1);
            } else {
                    System.err.println("Unknow type of tags!");
            }

    return tagValue;
}
/**
 * Accepts a string varName and returns the path that it
 * represents according to the variableTable.
 * If the path contains a VarStep, then this method will be recursively called
 * The end result will be the full path for a given column name.
 * 
 * @return NavigationStep[] The array containing all the variable's path.
 * @param varName The variable looking up.
 * @since 1.0
 */
public NavigationStep[] getVariableValue(String varName) {
	this.VarHash = new Hashtable();
	return getVariableValue(varName, true);
}
/**
 * Accepts a string varName and returns the path that it
 * represents according to the this.
 * If the path contains a VarStep, then this method will be recursively called
 * The end result will be the full path for a given column name.
 * 
 * @return NavigationStep[] containing the variable's path
 * @param varName containing the variable to look up.
 * @since 1.0
 */
public NavigationStep[] getVariableValue(
    String varName,
    boolean recursiveLookup) {

	    this.VarHash = new Hashtable();
	    
    return getVariableValueRecursive(varName, recursiveLookup);
}
/**
 * Accepts a string varName and returns the path that it
 * represents according to the this.
 * If the path contains a VarStep, then this method will be recursively called
 * The end result will be the full path for a given column name.
 * 
 * @return NavigationStep[] containing the variable's path
 * @param varName containing the variable to look up.
 * @since 1.0
 */
protected NavigationStep[] getVariableValueRecursive(
    String varName,
    boolean recursiveLookup) {
    NavigationStep[] returnStep;
    Vector resultVector = new Vector();

    if (varName == null) {
        return (NavigationStep[]) resultVector.toArray(new NavigationStep[] {
        });
    }

    // check for infinite loop.
    if (this.VarHash.get(varName) != null) {
        // found a circle call.
        System.err.println(
            "ERROR: BindingTable.getVariableValueRecursive(String, boolean) infinite loop for VarName:"
                + varName
                + "! Abort!");
        // reset the VarHash.
        this.VarHash = new Hashtable();
        // give back the unrecursive answer.
        return (NavigationStep[]) get(varName);
    } else {
        // put to the varhash.
        this.VarHash.put(varName, varName);
    }

    try {
        // if it is not recursive lookup, just return the value.
        // if we cannot find the key, the the result is NULL!
        if (!recursiveLookup) {
            // pop this varName from the hash table.
            this.VarHash.remove(varName);
            return (NavigationStep[]) get(varName);
        } else {
            // otherwise, it's a recursive lookup.

            // first, let's put the current's var's value into the result.
            NavigationStep[] result = (NavigationStep[]) this.get(varName);
            if (result == null) {
                // pop this varName from the hash table.
                this.VarHash.remove(varName);

                return null;
            }

            for (int j = 0; j < result.length; j++) {
                resultVector.addElement(result[j]);
            }

            // second, recursively iterate through the VarStep (if any) and extend it.
            for (int i = 0; i < resultVector.size(); i++) {
                //if there is a varStep 
                if ((NavigationStep) resultVector.elementAt(i) instanceof VarStep) {
                    VarStep match = (VarStep) resultVector.elementAt(i);
                    //then call this method again.  
                    NavigationStep[] variableStep =
                        getVariableValueRecursive((String) (match).getStepValue(), true);
                    //replace the varstep with its value.
                    int position = i;
                    resultVector.removeElement(match);

                    // if the variableStep has no value (null) break out
                    if (variableStep == null) {
                        continue;
                    }

                    for (int x = 0; x < variableStep.length; x++) {
                        //now have to insert veriableStep into resultVector, starting at position i
                        resultVector.insertElementAt(variableStep[x], position);
                        position++;

                    }

                }
            }
        }

        // pop this varName from the hash table.
        this.VarHash.remove(varName);
        // return the result.
        return (NavigationStep[]) resultVector.toArray(new NavigationStep[] {
        });

    } catch (ClassCastException ce) {

        System.out.println(ce.toString());
        System.out.println(
            "Tried to cast a "
                + this.get(varName).getClass().toString()
                + " into a NavigationStep[]");

        // pop this varName from the hash table.
        this.VarHash.remove(varName);
        // just return whatever we have before the Exception.
        return (NavigationStep[]) resultVector.toArray(new NavigationStep[] {
        });
    }

}
/**
 * Check if there is more content between lborder and rborder within given NavigationStep array.
 * @return boolean True: more content between. False: otherwise.
 * @param input The NavigationStep searching inside.
 * @param lborder The left border of searching.
 * @param rborder The right border of searching.
 * @since 1.0
 */
boolean moreContent(NavigationStep[] input, int lborder, int rborder) {
	if(input == null)
		return false;
	if(lborder < 0 ||
		rborder < 0 ||
		lborder > input.length-1 ||
		rborder > input.length-1)
		return false;
	else
		return rborder >= lborder;
}
/**
 * Convert a NavigationStep into String.
 * @return String The string represent the given Navigation Step.
 * @param steps The NavigationStep array converting for.
 * @since 1.0
 */
public String navStepsToString(NavigationStep[] steps) {
    StringWriter strWriter = new StringWriter();
    PrintWriter strOut = new PrintWriter(strWriter);

    if (steps != null) {
        for (int istep = 0; istep < steps.length; istep++) {
            if (steps[istep] instanceof AttStep)
                strOut.print("/@" + steps[istep]);
            else if (steps[istep] instanceof VarStep)
                strOut.print("/(" + steps[istep] + ")");
            else if (steps[istep] instanceof NodeStep)
                strOut.print("/." + steps[istep] + ".");
            else if (steps[istep] instanceof TagStep)
                strOut.print("/_" + steps[istep] + "_");
            else if (steps[istep] instanceof StringStep)
                strOut.print("/`" + steps[istep] + "'");
            else if (steps[istep] instanceof DescendantStep)
                strOut.print("/." + steps[istep] + ".");
            else
                strOut.print("/" + steps[istep]);

        }
    }

    return strWriter.toString();
}
/**
 * Get the next content between left and right border within the NavigationStep array.
 * @return NavigationStep[] The next NavigationStep array.
 * @param input The source NavigationStep searching inside.
 * @param lborder The left border of searching.
 * @param rborder The right border of searching.
 * @since 1.0
 */
NavigationStep[] nextContent(NavigationStep[] input, int lborder, int rborder) {
	ArrayList steps = new ArrayList();
	NavigationStep[] result = (NavigationStep[])steps.toArray(new NavigationStep[]{});

	// check the input parameters.
	if(input == null){
		steps.add(new NodeStepImp("ERROR"));
		return (NavigationStep[])steps.toArray(new NavigationStep[]{});
	}
	if(lborder < 0 ||
		rborder < 0 ||
		lborder > input.length-1 ||
		rborder > input.length-1){
		steps.add(new NodeStepImp("ERROR"));
		return (NavigationStep[])steps.toArray(new NavigationStep[]{});
	}

	// search for the next possible content, start from 0.

	// if that's a string.
	if(input[lborder] instanceof StringStep){
		steps.add(input[lborder]);
		result = (NavigationStep[])steps.toArray(new NavigationStep[]{});
	}else if(input[lborder] instanceof NodeStep){ // if that's a NodeStep.
		steps.add(input[lborder]);
		// search for the end of the NodeStep.
		int iStep = lborder + 1;
		while(iStep < input.length -1 && 
			input[iStep] instanceof NodeStep && 
			!input[iStep].getStepValue().equals("SOURCE")){
			steps.add(input[iStep]);
			iStep ++;
		}
		result = (NavigationStep[])steps.toArray(new NavigationStep[]{});
	}else if(input[lborder] instanceof TagStep){ // if that's a TagStep.
		steps.add(input[lborder]);
		// get the tag value.
		String tag = input[lborder].getStepValue();
		tag = tag.substring(1, tag.length() -1);
		// search for the end tag.
		int iStep = lborder + 1;
		while(iStep < input.length -1){
			// here we assume no existence of <a/>, but always <a></a>.
			steps.add(input[iStep]);
			if(input[iStep] instanceof TagStep){
				if(input[iStep].getStepValue().indexOf(tag) != -1){
					// if that's a match end tag.
					break;
				}
			}
			iStep ++;
		}
		result = (NavigationStep[])steps.toArray(new NavigationStep[]{});
	}else {// something tricky here.
		// It shouldn't reach here!
		System.err.println("WARNING: BindingTable:nextContent() SHOULDN'T BE HERE!");
		steps.add(new NodeStepImp("ERROR"));
		return (NavigationStep[])steps.toArray(new NavigationStep[]{});
	}
	
	return result;
}
/**
 * Put the variable binding and its key into the BindingTable.
 * @deprecated You should use updateVariableTable(..,..) instead.
 */
public Object put(Object key, Object value) {
    return super.put(key, value);
}
/**
 * Reduce the given NavigationStep array.
 * There are minor extension to handle the special navstep:
 * UNNEST and #text.
 * If we see the UNNEST, we delete this step.
 * The UNNEST doesn't affect the matching, and we will deal with the mismatch between 
 * NavUnnest and NavCollection at the Tree modification.
 * The #text will cancel out with any arbitrary tag if there is any.
 * @return The reduced NavigationStep array.
 * @param result The source NavigationStep array reducing.
 * @since 1.0
 */
private NavigationStep[] reduceSteps(NavigationStep[] result) {

    /* Preprocessing
    * Remove all the UNNEST step.
    */
    for (int istep = result.length - 1; istep >= 0; istep--) {
        if ("UNNEST".equals(result[istep].getStepValue()))
            result = removeContent(result, istep, 1);
    }

    /* Here is the algorithm:
    *  1. Find the last TagStep from the end.
    *     If there is no TagStep, then done.
    *  2. Check if there is a NodeStep after it.
    *     If there is no more, then done.
    *  3. Check if the TagStep same as the NodeStep.
    *     If they are same:
    *        delete the NodeStep. Goto step 2.
    *     If they are different goto step 4.
    *  4. Find the match open TagStep from the front.
    *     Delete both the open and close TagStep.
    *	  Do the distribution of the Tail.
    *     Iterate through the content.
    *     Locate the SOURCE NodeStep.
    *     Find the last NodeStep start with the SOURCE NodeStep.
    *     If that's not same as the to be matched NodeStep.
    *       delete that sequence.
    *     else keep that sequence.
    *     At then end, delete the to be matched NodeStep.
    *  5. goto step 2
    */
    // first step, locate the last TagStep. 
    int iTagStep = -1;
    for (int iStep = result.length - 1; iStep >= 0; iStep--) {
        if (result[iStep] instanceof TagStep) {
            iTagStep = iStep;
            break;
        }
    }

    // if there is no TagStep, do nothing.
    if (iTagStep <= -1) {
        return result;
    }

    // second step, find the next NodeStep.
    while (iTagStep < result.length - 1
        && result[iTagStep + 1] instanceof NodeStep) {
        // third step, check if the TagStep = NodeStep.
        String navStep = result[iTagStep + 1].getStepValue();

        String closeTag = result[iTagStep].getStepValue();
        closeTag = getTagValue(closeTag);
        if (closeTag.equals(navStep)) {
            // delete the navStep.
            result = removeContent(result, iTagStep + 1, 1);
        } else { // TagStep != NodeStep
            // forth step.
            // find a open Tag at the beginning.
            String openTag = result[0].getStepValue();
            openTag = getTagValue(openTag);
            // The open tag should be the first step.
            if (!closeTag.equals(openTag)) {
                // The open tag didn't match the close tag.
                System.err.println("ERROR: Open Tag does not match the close Tag.");
                return result;
            }
            // now the open tag match the close tag.

            // delete both the open tag and the close tag.
            result = removeContent(result, iTagStep, 1); // remove close tag.
            result = removeContent(result, 0, 1); // remove open tag.

            // NOW: The content to be checked is from 0 to iTagStep-2. The to-be-matched NavStep is at iTagStep-1.

            // distribute the content with the proper tales.
            // There are two kinds of content. First, a navigation. Second, another open Tag.
            // If that's a navigation, the to-be-matched Navigate should be same as the last step of this nagivate.
            // If that's a tagger, the tag should match the to-be-matched navigate.
            // Separate content is decided by SOURCE navigation step, a Tag step, or a Text step.
            int lborder = 0;
            int rborder = iTagStep - 2;
            while (moreContent(result, lborder, rborder)) {
                NavigationStep[] content = nextContent(result, lborder, rborder);
                boolean match = true;
                if (content[0].getStepValue().equals("SOURCE")) {
                    // that's a navigation step.
                    // test if the last step match the to-be-matched Step.
                    match = content[content.length - 1].getStepValue().equals(navStep);
                } else
                    if (content[0] instanceof TagStep) {
                        // that's an element construction.
                        // test if the tag match the to-be-matched Step.
                        openTag = content[0].getStepValue();
                        openTag = getTagValue(openTag);
                        match = navStep.equals(openTag);

                    } else
                        if (content[0] instanceof TextStep) {
                            // that's just a string.
                            // NOW, we remove it.
                            // But, we should only remove it, when the to-be-matched is not a TextStep.
                            // If the to-be-matched is a TextStep, then this content should be kept.
                            match = false;
                        } else {
                            // Something unexpected here.
                            // Might be an attribute or something.
                            // Don't want to consider now.
                            System.err.println(
                                "BindingTable:computeFullPath: WARNING: Meet an unexpected Type of content.");
                            // keep it.
                            match = true;
                        }
                if (match) {
                    // that's a match.
                    // keep this content.
                    lborder = lborder + content.length;

                    // ???distribute the tail, and recursively call the function???
                } else {
                    // if that's not a match.
                    // delete this content.
                    result = removeContent(result, lborder, content.length);
                    rborder = rborder - content.length;
                }
            }

            // remove the to-be-matched step.
            result = removeContent(result, rborder + 1, 1);

            // update the iTagStep
            iTagStep = rborder;

            // terminate the loop if iTagStep is less than zero, hence no TagStep anymore.
            // because, that means there is nothing to compare with.
            if (iTagStep < 0)
                break;
        }
    }

    /** Post-Processing
    * Remove all the #text if matched with a Tagger before it.
    */
    for (int istep = result.length - 1; istep >= 3; istep--) {
        if ("#text".equals(result[istep].getStepValue())) {
            // found a #text
            // check if the step before it is a Tagstep, that should be a close tag.
            if (result[istep - 1] instanceof TagStep) {
                String closetag = getTagValue(result[istep - 1].getStepValue());
                // find the open tag at the beginning of the string.
                String opentag = getTagValue(result[0].getStepValue());
                if (opentag.equals(closetag)) {
                    // find a match.
                    // erase the #text.
                    result = removeContent(result, istep, 1);
                    // erase the close tag.
                    result = removeContent(result, istep - 1, 1);
                    // erase the open tag.
                    result = removeContent(result, 0, 1);
                    // adjust the index.
                    istep = istep - 2;
                }

            } // else, don't do anything.
        }
    }
    return result;
}
/**
 * Remove a piece from a given NavigationSteps[], and return the updated array.
 * @return The NavigationStep[] after removing specified content.
 * @param input The source NavigationStep[] cutting.
 * @param index The start position.
 * @param length The length of cutting.
 * @since 1.0
 */
NavigationStep[] removeContent(NavigationStep[] input, int index, int length) {
	ArrayList steps = new ArrayList();
	NavigationStep[] result = (NavigationStep[])steps.toArray(new NavigationStep[]{});
	if(input == null || index < 0 || length < 0)
		return result;
	// keep the content before the index.
	for(int iStep = 0; iStep < index && iStep < input.length; iStep++)
		steps.add(input[iStep]);
	// keep the reminder.
	for(int iStep = index + length; iStep < input.length; iStep++)
		steps.add(input[iStep]);

	// return the result.
	result = (NavigationStep[])steps.toArray(new NavigationStep[]{});
		
	return result;
}
/**
 * Replace ALL the occurances of the oldKey with the newKey.
 * @return void
 * @param newKey The old key replaced.
 * @param oldKey The new key replacing.
 * @since 1.0
 */
public void replaceKey(String newKey, String oldKey) {

    // get the value for the old var.
    NavigationStep[] binding = getVariableValue(oldKey, false);
    /* now we will add in the new binding to the vartable*/
    this.put(newKey, binding);

    // remove the old key bindings.
    this.remove(oldKey);

    // replace all existance of the oldKey in all the values with the newKey.
    for (Enumeration e = this.elements(); e.hasMoreElements();) {

        NavigationStep[] value = (NavigationStep[]) e.nextElement();

        // iterate through the steps in the value and replace the VarStep if any.
        if (value == null)
            continue;

        for (int iStep = 0; iStep < value.length; iStep++) {
            NavigationStep curStep = value[iStep];
            if (curStep instanceof VarStep) {
                // compare the curStep with the oldKey.
                if (curStep.getStepValue().equals(oldKey)) {
                    // make it the new key.
                    curStep.setStepValue(newKey);
                }
            }
        }
    }

}
/**
 * Take a set of Navigation steps, and produce a XML path, minus text
 * This method is useful for getKeyByXMLValue().
 * @return The converted result string.
 * @param steps The source NavigationStep[].
 * @since 1.0
 */
public String simpleToXMLString(NavigationStep[] steps) {
    StringWriter strWriter = new StringWriter();
    PrintWriter strOut = new PrintWriter(strWriter);

    if (steps != null) {
        for (int istep = 0; istep < steps.length; istep++) {
	        if(!(steps[istep] instanceof TextStep)){
		        strOut.print("/" + steps[istep]);
	        }
        }
    }

    return strWriter.toString();
}
/**
 * Convert a ListPattern into an Array of NavigationStep.
 * @return The NavigationStep[] converted.
 * @param pattern The source ListPattern.
 * @since 1.0
 */
NavigationStep[] toNavigationStep(ListPattern pattern) {
    /* now we will add in the new binding to the vartable*/
    int size = pattern.getList().size();
    NavigationStep[] newBindingNav = new NavigationStepImp[size];
    for (int i = 0; i < size; i++) {
        Object temp = pattern.getList().elementAt(i);
        if (temp instanceof String) {
            String tag = (String) temp;
            NavigationStep tempStep = null;
            if (tag.length() > 2) {
                if (tag.charAt(1) == '/') { // check if that's a close tagger.
                    tempStep = new TagStepImp("</" + getTagValue(tag) + ">");
                } else if (tag.charAt(0) == '<') { // check if that's an open tagger.
                    tempStep = new TagStepImp("<" + getTagValue(tag) + ">");
                } else {
                    tempStep = new StringStepImp((String) temp);
                }
            } else {
                tempStep = new StringStepImp((String) temp);
            }
            newBindingNav[i] = tempStep;
        } else if (temp instanceof ColumnNameImp) {
            VarStep vartemp = new VarStepImp(((ColumnNameImp) temp).getColumnName());
            newBindingNav[i] = vartemp;
        }

    }
    return newBindingNav;
}
/**
 * Convert BindingTable into a String.
 * @return The converted result.
 * @since 1.0
 */
public String toString() {
	return toString(false);
}
/**
 * Output the String format of this BindingTable.
 * @return The converted result string.
 * @param fullpath True: output the fullpath of variable binding. False: otherwise.
 * @since 1.0
 */
public String toString(boolean fullpath) {
    // get all the keys.
    // iterate through them.
    // then, print it out in a beautiful format.
    StringWriter strWriter = new StringWriter();
    PrintWriter strOut = new PrintWriter(strWriter);

    // Get the iteration of the keys.
    Enumeration enum = this.keys();
    while (enum.hasMoreElements()) {
        // get the key.
        Object key = enum.nextElement();
        // get the values.
        NavigationStep[] steps = this.getVariableValue((String) key, fullpath);

        // print those out in the result.
        strOut.print(key + "\t\t|\t");
        strOut.print(navStepsToString(steps));
        strOut.println();
    }
    return strWriter.toString();
}
/**
 * Remove the entry for oldKey and make it bound to binding
 * 
 * @param newBinding The new variable binding for given key.
 * @param oldKey The key of the variable which will be replaced.
 * @since 1.0
 */
public void updateVariableTable(
    String oldKey,
    NavigationStep[] newBinding) {

    /* remove the existing binding for oldKey*/
    if (this.containsKey(oldKey))
        this.remove(oldKey);

    /* now we will add in the new binding to the vartable*/
    this.put(oldKey, newBinding);
}
/**
 * Put a listPattern into the binding table by converting the pattern into navigation steps.
 * ASSUMPTION: No consideration of the attribute here.
 * Otherwise, we have to removed this method, and use the DOMPattern directly.
 * It will persistent as the first-order traverse of the tree.
 * @param newBinding The new variable binding for given key.
 * @param oldKey The key of the variable which will be replaced.
 * @since 1.0
 */
public void updateVariableTable(
    String oldKey,
    ListPattern newBinding) {

    /* remove the existing binding for oldKey*/
    if (this.containsKey(oldKey))
        this.remove(oldKey);
    
    NavigationStep[] newBindingNav = toNavigationStep(newBinding);

	this.put(oldKey, newBindingNav);
}
/**
 * Make whatever is bound to the oldKey to be bound to the newKey
 * 
 * @param newkey The new variable binding for given key.
 * @param oldKey The key of the variable which will be replaced.
 * @since 1.0
 */
public void updateVariableTable(String newKey, String oldKey) {

    // get the value for the old var.
    NavigationStep[] binding = getVariableValue(oldKey, false);
    /* now we will add in the new binding to the vartable*/
    this.put(newKey, binding);

    // remove the old key bindings.
    this.remove(oldKey);

}
/**
 * Remove the entry for oldKey and make it bound to binding
 * 
 * @param newBinding The new variable binding for given key.
 * @param oldKey The key of the variable which will be replaced.
 * @since 1.0
 */
public void updateVariableTable(String oldKey, Vector newBinding) {

    /* remove the existing binding for oldKey*/
    if (this.containsKey(oldKey))
        this.remove(oldKey);

    /* now we will add in the new binding to the vartable*/
    int size = newBinding.size();
    NavigationStep[] newBindingNav = new NavigationStepImp[size];
    for (int i = 0; i < size; i++) {
        newBindingNav[i].setStepValue((String) newBinding.elementAt(i));
    }
    this.put(oldKey, newBindingNav);
}
}
