package edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer;

import java.util.Hashtable;
/**
 * This class contains information regarding statistics.  This is a utility class
 * for the statistics gatherer.  <p>
 *
 * Added 3/8: This class now outlines which properties are generic and which can be calculated 
 * for different categories.  This is part of the new version of the statistics gatherer where
 * non-category specific statistics (TUPLE_COUNT, NUMBER_OF_TUPLES_IN_INPUT_QUEUES. etc) are kept in
 * their own GENERIC category.  <p>
 * 
 * The StatisticsObject that is used for each property is contained here in <tt>initReferenceTable</tt>.  
 * Because new property value are calculated 
 * in different ways, we want a general way to figure out which StatisticsObject
 * to use when initilizing the StatisticsGather.  For example, when the property named
 * SELECTIVITY is encountered, we want to instantiate a certain StatisticsObject so that
 * SELECTIVITY is calculated correctly. <p>
 *
 * Whenever new properties are added (that require special calculation), the initReferenceTable()
 * method should be updated.
 * If a new property can use the default implementations of compareTo and calculateNewValue, then there
 * is no need to update this class.  <p>
 *
 * Remember, the default calculateNewValue method just takes the propertyValue from the operator
 * and stores it in the StatisticsObject <p>
 * 
 * This class also contains constants for extracting values in the ModifiedArrayList. <p>
 */
public class StatisticsReference implements java.io.Serializable {
	private Hashtable referenceTable;
	public static final int LAST = -1;
	public static final int HIGHEST = -2;
	public static final int LOWEST = -3;
	public static final int AVERAGE = -4;
	public static final int VALUE = -5;
	
	private static java.lang.String DEFAULT = "edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.StatisticsObject";
/**
 * StatisticsReference constructor comment.
 */
public StatisticsReference() {
	super();
	referenceTable = new Hashtable();
	initReferenceTable();
}
/**
 * Gets the name of the StatisticsObject that refers to this property.
 * If no StatisticsObject is found, return the default StatisticsObject
 * @return java.lang.String
 * @param propertyName java.lang.String
 */
public String getStatisticsObjectName(String propertyName) {
    /** first try the specific, then generic, then take the default */
    if (((Hashtable) referenceTable.get("SPECIFIC")).containsKey(propertyName)) {
        return (String) ((Hashtable) referenceTable.get("SPECIFIC")).get(propertyName);
    } else
        if (((Hashtable) referenceTable.get("GENERIC")).containsKey(propertyName)) {
            return (String) ((Hashtable) referenceTable.get("GENERIC")).get(propertyName);
        } else {
            return (String) ((Hashtable) referenceTable.get("GENERIC")).get("DEFAULT");
        }
}
/**
 * Initializes the referenceTable.
 * Updated 3/8: The reference table now is grouped by GENERIC and SPECIFIC (to a category) properties.
 * The referenceTable contains two entries, each of which are hashtables.  If a property is generic, that
 * is it common to all categories and would not be different for each category, it should be in the generic
 * category.  For example, TUPLE_COUNT (how many tuples are in a queue) is not specific to a category (like a scheduling
 * algorithm) and thus should be in generic.  However, a property like SELECTIVITY or GREEDY_PRIORITY would differ
 * depending on the category, therefore, it should be in specific. <p>
 *
 * The GENERIC statistics will be the same regardless of the category while the SPECIFIC stats are different for each. 
 * This method should be changed whenever a new
 * StatisticsObject or Property is added that requires its own StatisticsObject
 * If a new property is added that can use the default statisticsobject (calculates new value
 * by setting it to some new value), then there is no need to alter this method. <p>
 *
 * The key is the propertyName, the value is the full path name of the StatisticsObject
 * that provides the correct calculateNewValue method <p>
 */
private void initReferenceTable() {
	Hashtable generic = new Hashtable();
	Hashtable specific = new Hashtable();
	referenceTable.put("GENERIC", generic);
	referenceTable.put("SPECIFIC", specific);
	
	//put in the default
	generic.put("DEFAULT", this.DEFAULT);
	/** for the dummy operators */
	generic.put("REAL_SELECTIVITY", this.DEFAULT);
	generic.put("RESULT_DELAY", this.DEFAULT);
	/** for the rest */	
	//generic.put("TUPLE_COUNT", "edu.wpi.cs.dsrg.xmldb.xat.common.calculators.TupleCount");
	generic.put("TUPLE_COUNT", this.DEFAULT);
	generic.put("NUMBER_OF_TUPLES_IN_INPUT_QUEUES", "edu.wpi.cs.dsrg.xmldb.xat.common.calculators.NumberOfTuplesInInputQueues");
	generic.put("NUMBER_OF_TUPLES_IN_OUTPUT_QUEUES", "edu.wpi.cs.dsrg.xmldb.xat.common.calculators.NumberOfTuplesInOutputQueues");	
	generic.put("OLDEST_TUPLE_IN_QUEUE", "edu.wpi.cs.dsrg.xmldb.xat.common.calculators.OldestTupleInQueue");
	specific.put("NUMBER_OF_TIMES_RUN", "edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.IncrementValueByN");
	specific.put("TIME_TOOK_TO_RUN_TOTAL", "edu.wpi.cs.dsrg.xmldb.xat.common.calculators.TimeTookToRunTotal");
//	specific.put("NUMBER_OF_TUPLES_OUTPUTTED", "edu.wpi.cs.dsrg.xmldb.xat.common.calculators.NumberOfTuplesOutputted");
	specific.put("NUMBER_OF_TUPLES_OUTPUTTED", "edu.wpi.cs.dsrg.xmldb.xat.common.calculators.NumberOfTuplesOutputtedTotal");
//	specific.put("NUMBER_OF_TUPLES_OUTPUTTED_TOTAL", "edu.wpi.cs.dsrg.xmldb.xat.common.calculators.");
	specific.put("NUMBER_OF_TUPLES_DEQUEUED_TOTAL", "edu.wpi.cs.dsrg.xmldb.xat.common.calculators.NumberOfTuplesToDequeueTotal");
	specific.put("NUMBER_OF_POSSIBLE_OUTPUTS", "edu.wpi.cs.dsrg.xmldb.xat.common.calculators.NumberOfPossibleOutputs");	
	
	specific.put("SELECTIVITY", "edu.wpi.cs.dsrg.xmldb.xat.common.calculators.SelectivityCalculator");
	specific.put("AVERAGE_TUPLE_PROCESSING_TIME", "edu.wpi.cs.dsrg.xmldb.xat.common.calculators.AverageTupleProcessingTime");
	specific.put("GREEDY_PRIORITY", "edu.wpi.cs.dsrg.xmldb.xat.common.calculators.GreedyPriorityCalculator");	
	specific.put("AVERAGE_OUTPUT_RATE", "edu.wpi.cs.dsrg.xmldb.xat.common.calculators.OperatorOutputRateCalculator");	

	/** for trees */
	specific.put("OPERATOR_UTILIZATION", "edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.OperatorUtilization");
	specific.put("OUTPUT_RATE", "edu.wpi.cs.dsrg.xmldb.xat.common.calculators.TreeOutputRateCalculator");
	specific.put("THROUGHPUT", "edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.TreeThroughputCalculator");
	specific.put("NUMBER_OF_TIMES_OPERATORS_WERE_RUN", "edu.wpi.cs.dsrg.xmldb.xat.common.calculators.NumberOfTimesOperatorRunTotal");
	specific.put("TOTAL_TUPLES_IN_QUEUES", "edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.TotalTuplesInAllQueues");
	specific.put("AVERAGE_TUPLE_DELAY", "edu.wpi.cs.dsrg.xmldb.xat.common.calculators.AverageTupleDelay");
	
	/** for the rewrite engine */
	generic.put("NUMBER_OF_TIMES_RULES_APPLIED", "edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.IncrementValueByN");
	generic.put("TIME_SPENT_EVALUATING_RULES_TOTAL", "edu.wpi.cs.dsrg.xmldb.xat.component.statisticsgatherer.TimeSpentEvaluatingRulesTotal");			

	}
/**
 * Gets the name of the StatisticsObject that refers to this property.
 * If no StatisticsObject is found, return the default StatisticsObject
 * @return java.lang.String
 * @param propertyName java.lang.String
 */
public boolean isGenericProperty(String propertyName) {    
	return ((Hashtable) referenceTable.get("GENERIC")).containsKey(propertyName);   
}
}
